#pragma once
#include "openvr.h"
#include "version.h"
#include "public_vrtypes.h"
#include "vrtypes.h"
#include "vrannotation.h"
// ivrinput.h

namespace vr
{
namespace IVRInput_010
{
	// Maximum number of characters in an action name, including the trailing null
	static const uint32_t k_unMaxActionNameLength = 64;

	// Maximum number of characters in an action set name, including the trailing null
	static const uint32_t k_unMaxActionSetNameLength = 64;

	// Maximum number of origins for an action
	static const uint32_t k_unMaxActionOriginCount = 16;

	// Maximum number of characters in a bone name, including the trailing null
	static const uint32_t k_unMaxBoneNameLength = 32;

	enum EVRSkeletalTransformSpace
	{
		VRSkeletalTransformSpace_Model = 0,
		VRSkeletalTransformSpace_Parent = 1
	};

	enum EVRSkeletalReferencePose
	{
		VRSkeletalReferencePose_BindPose = 0,
		VRSkeletalReferencePose_OpenHand,
		VRSkeletalReferencePose_Fist,
		VRSkeletalReferencePose_GripLimit
	};

	enum EVRFinger
	{
		VRFinger_Thumb = 0,
		VRFinger_Index,
		VRFinger_Middle,
		VRFinger_Ring,
		VRFinger_Pinky,
		VRFinger_Count
	};

	enum EVRFingerSplay
	{
		VRFingerSplay_Thumb_Index = 0,
		VRFingerSplay_Index_Middle,
		VRFingerSplay_Middle_Ring,
		VRFingerSplay_Ring_Pinky,
		VRFingerSplay_Count
	};

	enum EVRSummaryType
	{
		// The skeletal summary data will match the animated bone transforms for the action.
		VRSummaryType_FromAnimation = 0,

		// The skeletal summary data will include unprocessed data directly from the device when available.
		// This data is generally less latent than the data that is computed from the animations.
		VRSummaryType_FromDevice = 1,
	};

	enum EVRInputFilterCancelType
	{
		VRInputFilterCancel_Timers = 0,
		VRInputFilterCancel_Momentum = 1,
	};

	enum EVRInputStringBits
	{
		VRInputString_Hand = 0x01,
		VRInputString_ControllerType = 0x02,
		VRInputString_InputSource = 0x04,

		VRInputString_All = 0xFFFFFFFF
	};

	struct InputAnalogActionData_t
	{
		/** Whether or not this action is currently available to be bound in the active action set */
		bool bActive;

		/** The origin that caused this action's current state */
		VRInputValueHandle_t activeOrigin;

		/** The current state of this action; will be delta updates for mouse actions */
		float x, y, z;

		/** Deltas since the previous call to UpdateActionState() */
		float deltaX, deltaY, deltaZ;

		/** Time relative to now when this event happened. Will be negative to indicate a past time. */
		float fUpdateTime;
	};

	struct InputDigitalActionData_t
	{
		/** Whether or not this action is currently available to be bound in the active action set */
		bool bActive;

		/** The origin that caused this action's current state */
		VRInputValueHandle_t activeOrigin;

		/** The current state of this action; will be true if currently pressed */
		bool bState;

		/** This is true if the state has changed since the last frame */
		bool bChanged;

		/** Time relative to now when this event happened. Will be negative to indicate a past time. */
		float fUpdateTime;
	};

	struct InputPoseActionData_t
	{
		/** Whether or not this action is currently available to be bound in the active action set */
		bool bActive;

		/** The origin that caused this action's current state */
		VRInputValueHandle_t activeOrigin;

		/** The current state of this action */
		TrackedDevicePose_t pose;
	};

	struct InputSkeletalActionData_t
	{
		/** Whether or not this action is currently available to be bound in the active action set */
		bool bActive;

		/** The origin that caused this action's current state */
		VRInputValueHandle_t activeOrigin;
	};

	struct InputOriginInfo_t
	{
		VRInputValueHandle_t devicePath;
		TrackedDeviceIndex_t trackedDeviceIndex;
		char rchRenderModelComponentName[128];
	};

	struct InputBindingInfo_t
	{
		char rchDevicePathName[128];
		char rchInputPathName[128];
		char rchModeName[128];
		char rchSlotName[128];
		char rchInputSourceType[ 32 ];
	};

	// * Experimental global action set priority *
	// These constants are part of the experimental support in SteamVR for overlay
	// apps selectively overriding input in the base scene application. This may be
	// useful for overlay applications that need to use part or all of a controller
	// without taking away all input to the game. This system must be enabled by the
	// "Experimental overlay input overrides" setting in the developer section of
	// SteamVR settings.
	//
	// To use this system, set the nPriority field of an action set to any number in
	// this range.
	static const int32_t k_nActionSetOverlayGlobalPriorityMin	= 0x01000000;
	static const int32_t k_nActionSetOverlayGlobalPriorityMax	= 0x01FFFFFF;

	static const int32_t k_nActionSetPriorityReservedMin		= 0x02000000;

	struct VRActiveActionSet_t
	{
		/** This is the handle of the action set to activate for this frame. */
		VRActionSetHandle_t ulActionSet;

		/** This is the handle of a device path that this action set should be active for. To
		* activate for all devices, set this to k_ulInvalidInputValueHandle. */
		VRInputValueHandle_t ulRestrictedToDevice;

		/** The action set to activate for all devices other than ulRestrictedDevice. If
		* ulRestrictedToDevice is set to k_ulInvalidInputValueHandle, this parameter is
		* ignored. */
		VRActionSetHandle_t ulSecondaryActionSet;

		// This field is ignored
		uint32_t unPadding;

		/** The priority of this action set relative to other action sets. Any inputs
		* bound to a source (e.g. trackpad, joystick, trigger) will disable bindings in
		* other active action sets with a smaller priority.
		*
		* Overlay applications (i.e. ApplicationType_Overlay) may set their action set priority
		* to a value between k_nActionSetOverlayGlobalPriorityMin and k_nActionSetOverlayGlobalPriorityMax
		* to cause any inputs bound to a source used by that action set to be disabled in scene applications.
		*
		* No action set priority may value may be larger than k_nActionSetPriorityReservedMin
		*/
		int32_t nPriority;
	};

	/** Contains summary information about the current skeletal pose */
	struct VRSkeletalSummaryData_t
	{
		/** The amount that each finger is 'curled' inwards towards the palm.  In the case of the thumb,
		* this represents how much the thumb is wrapped around the fist.
		* 0 means straight, 1 means fully curled */
		float	flFingerCurl[ VRFinger_Count ];

		/** The amount that each pair of adjacent fingers are separated.
		* 0 means the digits are touching, 1 means they are fully separated.
		*/
		float	flFingerSplay[ VRFingerSplay_Count ];
	};


	class IVRInput
	{
	public:

		// ---------------  Handle management   --------------- //

		/** Sets the path to the action manifest JSON file that is used by this application. If this information
		* was set on the Steam partner site, calls to this function are ignored. If the Steam partner site
		* setting and the path provided by this call are different, VRInputError_MismatchedActionManifest is returned.
		* This call must be made before the first call to UpdateActionState or IVRSystem::PollNextEvent. */
		virtual EVRInputError SetActionManifestPath( const char *pchActionManifestPath ) = 0;

		/** Returns a handle for an action set. This handle is used for all performance-sensitive calls. */
		virtual EVRInputError GetActionSetHandle( const char *pchActionSetName, VRActionSetHandle_t *pHandle ) = 0;

		/** Returns a handle for an action. This handle is used for all performance-sensitive calls. */
		virtual EVRInputError GetActionHandle( const char *pchActionName, VRActionHandle_t *pHandle ) = 0;

		/** Returns a handle for any path in the input system. E.g. /user/hand/right */
		virtual EVRInputError GetInputSourceHandle( const char *pchInputSourcePath, VRInputValueHandle_t *pHandle ) = 0;



		// --------------- Reading action state ------------------- //

		/** Reads the current state into all actions. After this call, the results of Get*Action calls
		* will be the same until the next call to UpdateActionState. */
		virtual EVRInputError UpdateActionState( VR_ARRAY_COUNT( unSetCount ) VRActiveActionSet_t *pSets, uint32_t unSizeOfVRSelectedActionSet_t, uint32_t unSetCount ) = 0;

		/** Reads the state of a digital action given its handle. This will return VRInputError_WrongType if the type of
		* action is something other than digital */
		virtual EVRInputError GetDigitalActionData( VRActionHandle_t action, InputDigitalActionData_t *pActionData, uint32_t unActionDataSize, VRInputValueHandle_t ulRestrictToDevice ) = 0;

		/** Reads the state of an analog action given its handle. This will return VRInputError_WrongType if the type of
		* action is something other than analog */
		virtual EVRInputError GetAnalogActionData( VRActionHandle_t action, InputAnalogActionData_t *pActionData, uint32_t unActionDataSize, VRInputValueHandle_t ulRestrictToDevice ) = 0;

		/** Reads the state of a pose action given its handle for the number of seconds relative to now. This
		* will generally be called with negative times from the fUpdateTime fields in other actions. */
		virtual EVRInputError GetPoseActionDataRelativeToNow( VRActionHandle_t action, ETrackingUniverseOrigin eOrigin, float fPredictedSecondsFromNow, InputPoseActionData_t *pActionData, uint32_t unActionDataSize, VRInputValueHandle_t ulRestrictToDevice ) = 0;

		/** Reads the state of a pose action given its handle. The returned values will match the values returned
		* by the last call to IVRCompositor::WaitGetPoses(). */
		virtual EVRInputError GetPoseActionDataForNextFrame( VRActionHandle_t action, ETrackingUniverseOrigin eOrigin, InputPoseActionData_t *pActionData, uint32_t unActionDataSize, VRInputValueHandle_t ulRestrictToDevice ) = 0;

		/** Reads the state of a skeletal action given its handle. */
		virtual EVRInputError GetSkeletalActionData( VRActionHandle_t action, InputSkeletalActionData_t *pActionData, uint32_t unActionDataSize ) = 0;

		/** Returns the current dominant hand for the user for this application. This function will only return success for applications
		* which include "supports_dominant_hand_setting": true in their action manifests. The dominant hand will only change after
		* a call to UpdateActionState, and the action data returned after that point will use the new dominant hand. */
		virtual EVRInputError GetDominantHand( ETrackedControllerRole *peDominantHand ) = 0;

		/** Sets the dominant hand for the user for this application. */
		virtual EVRInputError SetDominantHand( ETrackedControllerRole eDominantHand ) = 0;

		// ---------------  Static Skeletal Data ------------------- //

		/** Reads the number of bones in skeleton associated with the given action */
		virtual EVRInputError GetBoneCount( VRActionHandle_t action, uint32_t* pBoneCount ) = 0;

		/** Fills the given array with the index of each bone's parent in the skeleton associated with the given action */
		virtual EVRInputError GetBoneHierarchy( VRActionHandle_t action, VR_ARRAY_COUNT( unIndexArayCount ) BoneIndex_t* pParentIndices, uint32_t unIndexArayCount ) = 0;

		/** Fills the given buffer with the name of the bone at the given index in the skeleton associated with the given action */
		virtual EVRInputError GetBoneName( VRActionHandle_t action, BoneIndex_t nBoneIndex, VR_OUT_STRING() char* pchBoneName, uint32_t unNameBufferSize ) = 0;

		/** Fills the given buffer with the transforms for a specific static skeletal reference pose */
		virtual EVRInputError GetSkeletalReferenceTransforms( VRActionHandle_t action, EVRSkeletalTransformSpace eTransformSpace, EVRSkeletalReferencePose eReferencePose, VR_ARRAY_COUNT( unTransformArrayCount ) VRBoneTransform_t *pTransformArray, uint32_t unTransformArrayCount ) = 0;

		/** Reads the level of accuracy to which the controller is able to track the user to recreate a skeletal pose */
		virtual EVRInputError GetSkeletalTrackingLevel( VRActionHandle_t action, EVRSkeletalTrackingLevel* pSkeletalTrackingLevel ) = 0;

		// ---------------  Dynamic Skeletal Data ------------------- //

		/** Reads the state of the skeletal bone data associated with this action and copies it into the given buffer. */
		virtual EVRInputError GetSkeletalBoneData( VRActionHandle_t action, EVRSkeletalTransformSpace eTransformSpace, EVRSkeletalMotionRange eMotionRange, VR_ARRAY_COUNT( unTransformArrayCount ) VRBoneTransform_t *pTransformArray, uint32_t unTransformArrayCount ) = 0;

		/** Reads summary information about the current pose of the skeleton associated with the given action.   */
		virtual EVRInputError GetSkeletalSummaryData( VRActionHandle_t action, EVRSummaryType eSummaryType, VRSkeletalSummaryData_t * pSkeletalSummaryData ) = 0;

		/** Reads the state of the skeletal bone data in a compressed form that is suitable for
		* sending over the network. The required buffer size will never exceed ( sizeof(VR_BoneTransform_t)*boneCount + 2).
		* Usually the size will be much smaller. */
		virtual EVRInputError GetSkeletalBoneDataCompressed( VRActionHandle_t action, EVRSkeletalMotionRange eMotionRange, VR_OUT_BUFFER_COUNT( unCompressedSize ) void *pvCompressedData, uint32_t unCompressedSize, uint32_t *punRequiredCompressedSize ) = 0;

		/** Turns a compressed buffer from GetSkeletalBoneDataCompressed and turns it back into a bone transform array. */
		virtual EVRInputError DecompressSkeletalBoneData( const void *pvCompressedBuffer, uint32_t unCompressedBufferSize, EVRSkeletalTransformSpace eTransformSpace, VR_ARRAY_COUNT( unTransformArrayCount ) VRBoneTransform_t *pTransformArray, uint32_t unTransformArrayCount ) = 0;

		// --------------- Haptics ------------------- //

		/** Triggers a haptic event as described by the specified action */
		virtual EVRInputError TriggerHapticVibrationAction( VRActionHandle_t action, float fStartSecondsFromNow, float fDurationSeconds, float fFrequency, float fAmplitude, VRInputValueHandle_t ulRestrictToDevice ) = 0;

		// --------------- Action Origins ---------------- //

		/** Retrieve origin handles for an action */
		virtual EVRInputError GetActionOrigins( VRActionSetHandle_t actionSetHandle, VRActionHandle_t digitalActionHandle, VR_ARRAY_COUNT( originOutCount ) VRInputValueHandle_t *originsOut, uint32_t originOutCount ) = 0;

		/** Retrieves the name of the origin in the current language. unStringSectionsToInclude is a bitfield of values in EVRInputStringBits that allows the
			application to specify which parts of the origin's information it wants a string for. */
		virtual EVRInputError GetOriginLocalizedName( VRInputValueHandle_t origin, VR_OUT_STRING() char *pchNameArray, uint32_t unNameArraySize, int32_t unStringSectionsToInclude ) = 0;

		/** Retrieves useful information for the origin of this action */
		virtual EVRInputError GetOriginTrackedDeviceInfo( VRInputValueHandle_t origin, InputOriginInfo_t *pOriginInfo, uint32_t unOriginInfoSize ) = 0;

		/** Retrieves useful information about the bindings for an action */
		virtual EVRInputError GetActionBindingInfo( VRActionHandle_t action, VR_ARRAY_COUNT( unBindingInfoCount ) InputBindingInfo_t *pOriginInfo, uint32_t unBindingInfoSize, uint32_t unBindingInfoCount, uint32_t *punReturnedBindingInfoCount ) = 0;

		/** Shows the current binding for the action in-headset */
		virtual EVRInputError ShowActionOrigins( VRActionSetHandle_t actionSetHandle, VRActionHandle_t ulActionHandle ) = 0;

		/** Shows the current binding all the actions in the specified action sets */
		virtual EVRInputError ShowBindingsForActionSet( VR_ARRAY_COUNT( unSetCount ) VRActiveActionSet_t *pSets, uint32_t unSizeOfVRSelectedActionSet_t, uint32_t unSetCount, VRInputValueHandle_t originToHighlight ) = 0;

		/** Use this to query what action on the component returned by GetOriginTrackedDeviceInfo would trigger this binding. */
		virtual EVRInputError GetComponentStateForBinding( const char *pchRenderModelName, const char *pchComponentName,
			const InputBindingInfo_t *pOriginInfo, uint32_t unBindingInfoSize, uint32_t unBindingInfoCount,
			RenderModel_ComponentState_t *pComponentState ) = 0;


		// --------------- Legacy Input ------------------- //
		virtual bool IsUsingLegacyInput() = 0;


		// --------------- Utilities ------------------- //

		/** Opens the binding user interface. If no app key is provided it will use the key from the calling process.
		* If no set is provided it will open to the root of the app binding page. */
		virtual EVRInputError OpenBindingUI( const char* pchAppKey, VRActionSetHandle_t ulActionSetHandle, VRInputValueHandle_t ulDeviceHandle, bool bShowOnDesktop ) = 0;

		/** Returns the variant set in the current bindings. If the binding doesn't include a variant setting, this function
		* will return an empty string */
		virtual EVRInputError GetBindingVariant( VRInputValueHandle_t ulDevicePath,
			VR_OUT_STRING() char *pchVariantArray, uint32_t unVariantArraySize ) = 0;

	};

	static const char * const IVRInput_Version = "IVRInput_010";

} // namespace vr

} // Close custom namespace
